unit instrMultiFileU;

{$include hinstr.inc}

interface

uses
  Classes,
  types,
  fpjson,
  jsonparser,
  LazFileUtils,
  instrPiperU;

type

  { TMultiFileInstr }

  TMultiFileInstr = class
  public
    Files: TStringDynArray;
    KeepPreInstrFiles: Boolean;
    Reversed: Boolean;
    constructor Create;
    procedure Instrumentate(aFilePath: string);
    procedure Instrumentate(aInputFilePath, aOutputFilePath: string);
    procedure Reverse(aFilePath: string);
    procedure Go;
    function GetInstrFilePath(aFilePath: string): string;
    function GetPreFilePath(aFilePath: string): string;
    function LoadSettings(aFilePath: string): Boolean;
    procedure LoadSettings(aJson: TJSONData);
    destructor Destroy; override;
  end;

implementation

{ TMultiFileInstr }

constructor TMultiFileInstr.Create;
begin
  inherited Create;
  KeepPreInstrFiles := True;
end;

procedure TMultiFileInstr.Instrumentate(aFilePath: string);
const
  DebuggingOutputEnabled: Boolean = True;
var
  instrFilePath, preFilePath: string;
begin
  if DebuggingOutputEnabled then
    WriteLN('Now instrumentating file: "' + aFilePath + '"...');
  instrFilePath := GetInstrFilePath(aFilePath);
  preFilePath := GetPreFilePath(aFilePath);
  Instrumentate(aFilePath, instrFilePath);
  RenameFileUTF8(aFilePath, preFilePath);
  RenameFileUTF8(instrFilePath, aFilePath);
  if not KeepPreInstrFiles then
    DeleteFileUTF8(preFilePath);
end;

procedure TMultiFileInstr.Instrumentate(aInputFilePath, aOutputFilePath: string);
var
  piper: TPiper;
begin
  piper := TPiper.Create;
  piper.OpenInputFile(aInputFilePath);
  piper.OpenOutputFile(aOutputFilePath);
  piper.Go;
  piper.Free;
end;

procedure TMultiFileInstr.Reverse(aFilePath: string);
var
  preFilePath: string;
begin
  preFilePath := GetPreFilePath(aFilePath);
  if FileExistsUTF8(preFilePath) then
  begin
    DeleteFileUTF8(aFilePath);
    RenameFileUTF8(preFilePath, aFilePath);
  end;
end;

procedure TMultiFileInstr.Go;
var
  i: Integer;
begin
  for i := 0 to Length(Files) - 1 do
  begin
    if not Reversed then
    begin
      Instrumentate(Files[i]);
      WriteLN(i + 1, ' of ', Length(Files), ' files instrumentated.');
    end
    else
    begin
      Reverse(Files[i]);
      WriteLN(i + 1, ' of ', Length(Files), ' files reversed.');
    end;
  end;
end;

function TMultiFileInstr.GetInstrFilePath(aFilePath: string): string;
begin
  result := aFilePath + '.instr';
end;

function TMultiFileInstr.GetPreFilePath(aFilePath: string): string;
begin
  result := aFilePath + '.preinstr';
end;

function TMultiFileInstr.LoadSettings(aFilePath: string): Boolean;
var
  f: TFileStream;
  json: TJSONData;
  parser: TJSONParser;
begin
  if FileExistsUTF8(aFilePath) then
  begin
    f := TFileStream.Create(aFilePath, fmOpenRead);
    parser := TJSONParser.Create(f);
    json := parser.Parse;
    parser.Free;
    LoadSettings(json);
    json.Free;
    result := True;
  end
  else
    result := False;
end;

procedure TMultiFileInstr.LoadSettings(aJson: TJSONData);
var
  o: TJSONObject;
  lFiles: TJSONArray;
  i: Integer;
begin
  o := aJson as TJSONObject;
  lFiles := o.Get('files', TJSONArray(nil));
  SetLength(Files, lFiles.Count);
  for i := 0 to Length(Files) - 1 do
    Files[i] := lFiles[i].AsString;
  KeepPreInstrFiles := o.Get('keepPreInstrFiles', True);
end;

destructor TMultiFileInstr.Destroy;
begin
  inherited Destroy;
end;

end.

