unit instrPiperU;

{$include hinstr.inc}

interface

uses
  types, Classes, sysutils,
  Console,
  PStr_ContSearcherU,
  StringUtilU,
  StringBuilderU;

type

  { TPiper }

  TPiper = class
  protected
    FIncoming: TStream;
    FOutgoing: TStream;
    {}
    InsideCBComment: Boolean;
    // Got first char of slash-slash comment;
    GFCOSSC: Boolean;
    //
    InsideSSComment: Boolean;
    (* Got first char of brace star comment; *)
    GFCOBSC: Boolean;
    (**)
    InsideBSComment: Boolean;
    (* Got first char of brace star comment end; *)
    GFCOBSCE: Boolean;
    // Just entered comment;
    JEC: Boolean;
    InsideString: Boolean;
    InsideImplementation: Boolean;
    Searcher: TMultiSearcher;
    UsesBlockUpdated: Boolean;
    CurrentProcedureNameBuilder: TStringBuilder;
    ProcedureName: TStringDynArray;
    CodeBlockLevel: Integer;
    FileUnitName: string;
    JH13: Boolean;
    LineNumber: Integer;
    FaceMode: Boolean;
    function GetInsideComment: Boolean;
    function NextChar: Char;
    procedure Go(c: Char);
    procedure GoNotComment(c: Char);
    procedure GoTestWrite(c: Char);
    procedure Write(c: Char);
    procedure Write(s: string);
    procedure Rewind(s: string);
    procedure CheckEnteringComment(c: Char);
    procedure CheckExitingComment(c: Char);
    procedure BeginDoingNothing;
    procedure ReadProcedureName(c: Char);
    function ImproveProcedureName(s: string): string;
    procedure HandleUsesBlock;
    // Potential pitfall: if there is no uses block in the unit, the unit won't be inserted.
    procedure UpdateUsesBlock;
    procedure UpdateBegin;
    procedure UpdateEnd;
    procedure HandleBegin;
    procedure HandleEnd;
    function GetArgs: string;
    procedure PushProcedureName(a: string);
    function GetCurrentProcedureName: string;
    function PopCurrentProcedureName: string;
    procedure ClearProcedureName;
    property InsideComment: Boolean read GetInsideComment;
    property CurrentProcedureName: string read GetCurrentProcedureName;
    procedure HandleExit;
    procedure UpdateExit;
  public
    constructor Create;
    procedure OpenInputFile(aFilePath: string);
    procedure OpenOutputFile(aFilePath: string);
    procedure Go;
    destructor Destroy; override;
  end;

const
  Quote = '''';
  MaxNestedProcs = 128;
  InstrumentationUnitName = 'InstrationU';

implementation

type
  TSillyClass = class
  end;

{ TPiper }

{ Some commented code. Do not remove; needed for testing.
constructor TPiper.Create;
begin
  inherited Create;
  Searcher := TMultiSearcher.Create;
  Searcher.Whole := True;
  Searcher.IgnoreCase := True;
  Searcher.Setup(ToDynArray(['begin', 'end', 'function', 'procedure', 'uses']));
end;
}

constructor TPiper.Create;
begin
  inherited Create;
  Searcher := TMultiSearcher.Create;
  Searcher.Whole := True;
  Searcher.IgnoreCase := True;
  Searcher.Setup(
    ToDynArray(
      ['begin',
      'end',
      'function',
      'procedure',
      'constructor',
      'destructor',
      'uses',
      'implementation',
      'case',
      'external',
      'asm',
      'assembler',
      'exit',
      'try']
    )
  );
  CodeBlockLevel := 0;
  SetLength(ProcedureName, MaxNestedProcs);
end;

procedure TPiper.OpenInputFile(aFilePath: string);
begin
  FIncoming := TFileStream.Create(aFilePath, fmOpenRead);
  FileUnitName := ChangeFileExt(ExtractFileName(aFilePath), '');
end;

procedure TPiper.OpenOutputFile(aFilePath: string);
begin
  FOutgoing := TFileStream.Create(aFilePath, fmCreate or fmOpenWrite);
end;

procedure TPiper.Go;
var
  c: Char;
begin
  LineNumber := 1;
  while True do
  begin
    c := NextChar;
    if #0 = c then
      break;
    Go(c);
  end;
end;

destructor TPiper.Destroy;
begin
  FIncoming.Free;
  FOutgoing.Free;
  Searcher.Free;
  inherited Destroy;
end;

procedure TPiper.CheckEnteringComment(c: Char);
begin
  {}
  if c = '{' then
    InsideCBComment := True;
  //
  if c = '/' then
  begin
    if GFCOSSC then
    begin
      InsideSSComment := True;
      GFCOSSC := False;
    end
    else
      GFCOSSC := True;
  end
  else
    GFCOSSC := False;
  (**)
  if GFCOBSC and (c = '*') then
  begin
    InsideBSComment := True;
    GFCOBSC := False;
  end;
  if c = '(' then
    GFCOBSC := True
  else
    GFCOBSC := False;
  // 'str'
  if c = Quote then
    InsideString := True;
end;

procedure TPiper.CheckExitingComment(c: Char);
begin
  {}
  if InsideCBComment and (c = '}') then
    InsideCBComment := False;
  //
  if InsideSSComment and (c = #10) then
    InsideSSComment := False;
  if InsideSSComment and (c = #13) then
    InsideSSComment := False;
  (**)
  if InsideBSComment and (c = '*') then
    GFCOBSCE := True;
  if InsideBSComment and (c = ')') and GFCOBSCE then
    InsideBSComment := False;
  // 'str'
  if InsideString and (c = Quote) then
    InsideString := False;
end;

procedure TPiper.BeginDoingNothing;

procedure SillyNestedProcedure;
begin
end;

BEgin
enD;

procedure TPiper.ReadProcedureName(c: Char);
const
  DebuggingOutputEnabled: Boolean = False;
begin
  if CurrentProcedureNameBuilder <> nil then
  begin
    if (c = ' ') or (c = #10) or (c = #13) then
      // skip
    else if (c = '(') and (CurrentProcedureNameBuilder.TotalLength = 0) then // type procedure
      FreeAndNil(CurrentProcedureNameBuilder)
    else if c = ';' then
    begin
      PushProcedureName(ImproveProcedureName(CurrentProcedureNameBuilder.ToString));
      CurrentProcedureNameBuilder.Free;
      CurrentProcedureNameBuilder := nil;
      if DebuggingOutputEnabled then
        WriteLN('Got procedure name: ' + CurrentProcedureName, ' ', CodeBlockLevel);
    end
    else
      CurrentProcedureNameBuilder.Add(c);
  end;
end;

function TPiper.ImproveProcedureName(s: string): string;
var
  i: Integer;
  c: Char;
begin
  for i := Length(s) downto 1 do
  begin
    c := s[i];
    if c = #10 then
      Delete(s, i, 1);
    if c = #13 then
      Delete(s, i, 1);
  end;
  result := s;
end;

procedure TPiper.HandleUsesBlock;
begin
  if not UsesBlockUpdated then
  begin
    UsesBlockUpdated := True;
    UpdateUsesBlock;
  end;
end;

procedure TPiper.UpdateUsesBlock;
begin
  Write(' ' + InstrumentationUnitName + ',');
end;

procedure TPiper.UpdateBegin;
begin
  Write(' instration_begin(' + GetArgs + ');');
end;

procedure TPiper.UpdateEnd;
begin
  if not FaceMode then
  begin
    Rewind('end');
    Write('{end};instration_end(' + GetArgs + '); end');
  end;
end;

procedure TPiper.HandleBegin;
begin
  if CodeBlockLevel = 0 then
    if CurrentProcedureName <> '' then
      UpdateBegin;
  Inc(CodeBlockLevel);
end;

procedure TPiper.HandleEnd;
begin
  if CodeBlockLevel > 0 then
  begin
    Dec(CodeBlockLevel);
    if CodeBlockLevel = 0 then
    begin
      if CurrentProcedureName <> '' then
        UpdateEnd;
      PopCurrentProcedureName;
    end;
  end
  else
    ClearProcedureName;
end;

function TPiper.GetArgs: string;
begin
  result :=
    QuotedStr(CurrentProcedureName) + ', '
    + QuotedStr(FileUnitName) + ', '
    + IntToStr(LineNumber)
  ;
end;

procedure TPiper.PushProcedureName(a: string);
var
  i: Integer;
begin
  for i := 0 to Length(ProcedureName) - 1 do
    if ProcedureName[i] = '' then
    begin
      ProcedureName[i] := a;
      break;
    end;
end;

function TPiper.GetCurrentProcedureName: string;
var
  i: Integer;
  first: Boolean;
begin
  result := '';
  first := True;
  for i := 0 to Length(ProcedureName) - 1 do
    if ProcedureName[i] <> '' then
    begin
      if not first then
        result := result + '.';
      result := result + ProcedureName[i];
      if first then
        first := False;
    end
    else
      break;
end;

function TPiper.PopCurrentProcedureName: string;
var
  i: Integer;
begin
  for i := 1 to Length(ProcedureName) - 1 do
    if ProcedureName[i] = '' then
    begin
      result := ProcedureName[i - 1];
      ProcedureName[i - 1] := '';
      break;
    end;
end;

procedure TPiper.ClearProcedureName;
var
  i: Integer;
begin
  for i := 0 to Length(ProcedureName) - 1 do
    ProcedureName[i] := '';
end;

procedure TPiper.HandleExit;
begin
  UpdateExit;
end;

procedure TPiper.UpdateExit;
begin
  Rewind('exit');
  Write('begin instration_end(' + GetArgs + '); exit end');
end;

function TPiper.GetInsideComment: Boolean;
begin
  result := InsideCBComment or InsideSSComment or InsideBSComment or InsideString;
end;

function TPiper.NextChar: Char;
begin
  result := #0;
  FIncoming.Read(result, SizeOf(result));
end;

procedure TPiper.Go(c: Char);
begin
  if (c = #10) and not (JH13) then
    Inc(LineNumber);
  if c = #13 then
  begin
    Inc(LineNumber);
    JH13 := True;
  end
  else
    JH13 := False;
  JEC := InsideComment;
  if not InsideComment then
    CheckEnteringComment(c);
  JEC := JEC <> InsideComment;
  if not InsideComment then
    GoNotComment(c);
  Write(c);
  if InsideComment and not JEC then
    CheckExitingComment(c);
end;

procedure TPiper.GoNotComment(c: Char);
const
  DebuggingOutputEnabled: Boolean = True;
var
  found: string;
begin
  found := Searcher.Match(c);
  if found = 'uses' then
    HandleUsesBlock;
  ReadProcedureName(c);
  if (found = 'procedure') and (c <> '(')
  or (found = 'function') and (c <> '(')
  or (found = 'constructor')
  or (found = 'destructor')
  then
    if InsideImplementation then
    begin
      CurrentProcedureNameBuilder := TStringBuilder.Create;
    end;
  if found = 'implementation' then
  begin
    InsideImplementation := True;
    CodeBlockLevel := 0;
  end;
  if (found = 'case') and (InsideImplementation) and (CurrentProcedureName <> '') then
    Inc(CodeBlockLevel);
  if (found = 'asm') then
    Inc(CodeBlockLevel);
  if found = 'begin' then
    HandleBegin;
  if found = 'end' then
    HandleEnd;
  if (found = 'external') or (found = 'assembler') then
    PopCurrentProcedureName;
  if found = 'exit' then
    HandleExit;
  if found = 'try' then
    Inc(CodeBlockLevel);
end;

procedure TPiper.GoTestWrite(c: Char);
begin
  if InsideComment then
    TextColor(Console.LightGray)
  else
    TextColor(Console.LightRed);
  System.Write(c);
end;

procedure TPiper.Write(c: Char);
begin
  FOutgoing.Write(c, SizeOf(c));
end;

procedure TPiper.Write(s: string);
begin
  FOutgoing.Write(s[1], Length(s));
end;

procedure TPiper.Rewind(s: string);
begin
  FOutgoing.Seek(0 - Length(s), soFromCurrent);
end;

end.

